#!/usr/bin/env python

import sys
import os
import os.path
import atexit
import tempfile
import shutil
import subprocess as sub
import threading
import signal
import resource
import time
import glob

root = os.path.dirname (os.path.dirname (os.path.realpath (__file__)))
verbose = True
running_process = None

os.environ['PATH'] = os.path.join (root, 'bin') + os.pathsep + os.environ['PATH']

def isexec (fpath):
    if fpath == None: return False
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

def which(program):
    fpath, fname = os.path.split(program)
    if fpath:
        if isexec (program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if isexec (exe_file):
                return exe_file
    return None

def kill (proc):
    try:
        proc.terminate ()
        proc.kill ()
        proc.wait ()
        global running_process
        running_process = None
    except OSError:
        pass

def cat (in_file, out_file): out_file.write (in_file.read ())


def createWorkDir (dname = None, save = False):
    if dname == None:
        workdir = tempfile.mkdtemp (prefix='ufo-')
    else:
        workdir = dname

    if verbose:
        print "Working directory", workdir

    if not save:
        atexit.register (shutil.rmtree, path=workdir)
    return workdir

def getLlbmc ():
    llbmc = which ("llbmc")
    if llbmc is None:
        raise IOError ("Cannot find llbmc")
    return llbmc

def getOpt26 ():
    v = which ('opt-2.6')
    if v is None:
        raise IOError ('Cannot find opt-2.6')
    return v

def getLlvmDis29 ():
    v = which ('llvm-dis-2.9')
    if v is None: raise IOError ('Cannot find llvm-dis-2.9')
    return v

def getLlvmAs32 ():
    v = which ('llvm-as-3.2')
    if v is None: raise IOError ('Cannot find llvm-as-3.2')
    return v

def killall ():
    global running_process
    if running_process != None:
        running_process.terminate ()
        running_process.kill ()
        running_process.wait ()
        running_process = None

def sharedLib (base):
    ext = '.so'
    if sys.platform.startswith ('darwin'): ext = '.dylib'
    return base + ext

def toLlvm32 (oldbc, newbc):

    ufofeLib = sharedLib ('lib/libUfoFe')
    ufofeLib = os.path.join (root, ufofeLib)

    optArg = [getOpt26 (), '-load', ufofeLib, '-kill-nondet', '-raiseallocs', '-fixll']
    disArg = [getLlvmDis29 ()]
    asmArg = [getLlvmAs32 (), '-o', newbc]

    if verbose:
        print ' '.join (optArg), '<', oldbc, '|'
        print ' '.join (disArg), '|'
        print ' '.join (asmArg)

    opt = sub.Popen (optArg, stdin=open (oldbc), stdout=sub.PIPE)
    dis = sub.Popen (disArg, stdin=opt.stdout, stdout=sub.PIPE)
    asm = sub.Popen (asmArg, stdin=dis.stdout, stdout=sub.PIPE)
    dis.stdout.close ()
    output = asm.communicate () [0]

    if asm.returncode != 0:
        raise sub.CalledProcessError (asm.returncode, asmArg)

def is_llbmc_opt (x):
    llbmc_opt = ['max-builtins-iterations', \
                 'max-function-call-depth', \
                 'max-loop-iterations', \
                 'report', \
                 'result', \
                 'counterexample', \
                 'no-max-loop-iterations-checks', \
                 'ignore-undetermined-functions', \
                 'ignore-missing-function-bodies', \
                 'ignore-inline-assembly',
                 'uninitialized-globals' ]
    if x.startswith ('-'):
        y = x.strip ('-')
        return y in llbmc_opt
    return False

def llbmc_result (f):
    if f is None: return None

    line = ""
    lines = open (f, 'rb')
    for l in lines : line = l
    if verbose: print 'LLBMC:', line
    if line.startswith ('Error detected.'): return 1
    #if line.startswith ('No error detected.') : return 0
    return 2 # i.e. UNKNOWN

def llbmc (in_file, loop_bound=0, args = [], stdout_name=None, cpu = -1, mem = -1):
    def set_limits ():
        if mem > 0:
            mem_bytes = mem * 1024 * 1024
            resource.setrlimit (resource.RLIMIT_AS, [mem_bytes, mem_bytes])
    llbmc_args = [getLlbmc (),  '-only-custom-assertions', \
                  '-no-max-loop-iterations-checks', \
                  '-ignore-missing-function-bodies', \
                  '-enforce-max-loop-iterations', \
                  '-ignore-inline-assembly', \
                  '-ignore-undetermined-functions', \
                  '-counterexample', \
                  '-max-loop-iterations={}'.format(loop_bound), \
                  in_file]
    llbmc_args.extend (args)
    llbmc_args.append ('-result')
    if verbose: print ' '.join (llbmc_args)

    stdout=None
    if stdout_name <> None: stdout = open (stdout_name, 'wb')

    p = sub.Popen (llbmc_args, preexec_fn=set_limits, stdout=stdout)
    global running_process
    running_process = p

    timer = threading.Timer (cpu, kill, [p])
    if cpu > 0: timer.start ()

    try:
        (pid, returnvalue, ru_child) = os.wait4 (p.pid, 0)
        running_process = None
    except OSError:
        pass
    finally:
        ## kill the timer if the process has terminated already
        if timer.isAlive (): timer.cancel ()

    return llbmc_result (stdout_name)



def nothing (x, y):
    pass
